package ch.epfl.tpp;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Stack;

public class DAG<N> {
	private Set<N> nodes = new HashSet<>();
	private Set<Edge<N>> edges = new HashSet<>();
	
	public void addNode(N n) {
		nodes.add(n);
	}

	public void removeNode(N n) {
		Iterator<Edge<N>> edgeIt = edges.iterator();
        while (edgeIt.hasNext()) {
            Edge<N> edge = edgeIt.next();
            if (n.equals(edge.fromNode) || n.equals(edge.toNode))
                edgeIt.remove();
        }
		nodes.remove(n);
	}

	public void addEdge(N f, N t) {
		if (! (nodes.contains(f) && nodes.contains(t)))
			throw new IllegalArgumentException();
		if (isReachable(t, f))
			throw new IllegalArgumentException();
		edges.add(new Edge<N>(f, t));
	}
	
	public Set<N> successors(N n) {
	    if (!nodes.contains(n))
	        throw new IllegalArgumentException();
		Set<N> successors = new HashSet<N>();
		for( Edge<N> edge : edges ){
			if( n.equals( edge.fromNode ) )
				successors.add(edge.toNode);
		}
		return successors;
	}

	public boolean isReachable(N f, N t) {
	    Stack<N> toVisit = new Stack<>();
	    toVisit.push(f);
	    while (!toVisit.isEmpty()) {
	        N n = toVisit.pop();
	        if (n.equals(t))
	            return true;
	        for (N s: successors(n))
	            toVisit.push(s);
	    }
	    return false;
	}
}

class Edge<N> {
    final N fromNode, toNode;
    
    public Edge(N fromNode, N toNode) {
        assert fromNode != null && toNode != null;
        this.fromNode = fromNode;
        this.toNode = toNode;
    }
    
    @Override
    public int hashCode() {
        return fromNode.hashCode() + 31 * toNode.hashCode();
    }

    private boolean equalsEdge(Edge<?> that) {
        return this.fromNode.equals(that.fromNode) && this.toNode.equals(that.toNode);
    }
    
    @Override
    public boolean equals(Object o) {
        return (o instanceof Edge<?>) && equalsEdge((Edge<?>)o);
    }
}
